/* dooralarm.c - firmware for door alarm 

		LED	HORN 	RFID
		O	O	I		
                               
	8	7	6	5
     +-------------------------------+	
     |	Vcc	PB2	PB1	PB0  |
      \                              |
       |      ATtiny 45              |
      /                              |
     |	RESET	pb3	pb4	gnd  |
     +-------------------------------+
	1 	2	3	4

		I	I 	
		DOOR	DISARM


		TLE4905L                                -------- Q
	    +--------------+                           /
	    |   \      /   |                         |/        
            |    |    |    |   MAGNETIC FIELD >------|
            |    |    |    |                         |\
            |    |    |    |                           Y
            |    |    |    |                            --------- GND
	    |   /      \   |
	    +--------------+   "magnetic field controlled transistor" 
	       1    2    3
              Vcc  GND   Q

	0 -> magnetic field detected 
	1 -> no magnetic field (via pull up)

*/

#define F_CPU 8000000UL

#include <avr/interrupt.h>
#include <util/delay.h> 

#define u8 unsigned char

u8 ready();
u8 arming();
u8 arming_lock();
u8 armed();
u8 disarming();
u8 disarming_lock();
u8 trigger_stage_00();
u8 trigger_stage_01(); 
u8 trigger_stage_02(); 
u8 trigger_stage_03();
u8 trigger_stage_04();
u8 trigger_stage_05();
u8 trigger_stage_06();
u8 trigger_stage_07();
u8 trigger_stage_08();
u8 trigger_stage_09();
u8 trigger_stage_10();
u8 trigger_stage_11();
u8 trigger_stage_12();
u8 trigger_stage_13();
u8 trigger_stage_14();
u8 trigger_stage_15();
u8 trigger_stage_16();
u8 trigger_stage_17();
u8 trigger_stage_18();
u8 trigger_stage_19();
u8 trigger_stage_20();

enum alarm_state {
	ALARM_ON = 1,
	ALARM_OFF = 2
};

enum led_state {
	LED_RED = 1,
	LED_GREEN = 2
};

enum rfid_state {

	RFID_ON = 1,
	RFID_OFF = 2
};

enum door_state {
	DOOR_OPEN = 1,
	DOOR_CLOSED = 2
};

enum alarm_override_state {
	ALARM_ENABLED = 1,
	ALARM_DISABLED = 2
};

#define TRANSITION(COND,NEXT_STATE)  if (COND){ state = &NEXT_STATE; return 1;}

/* system power on -> system is armed (prevent disarm by short power interruption) */
u8 (*state)() = &armed;
enum alarm_state alarm = ALARM_ON;
enum led_state led = LED_RED;

unsigned int timer_c = 0; /* counter used for timer interval */

/* get external states */

enum rfid_state getRFIDState() 
{
	return  (PINB & (1<<PB0)) != 0 ? RFID_ON : RFID_OFF;
}

enum door_state getDoorState() 
{
	return  (PINB & (1<<PB3)) != 0 ? DOOR_OPEN : DOOR_CLOSED;
}

enum alarm_override_state getAlarmOverrideState() 
{
	return  (PINB & (1<<PB4)) != 0 ? ALARM_ENABLED : ALARM_DISABLED;
}


/* the alarm system states */

u8 ready() 
{
	TRANSITION( getRFIDState() == RFID_ON , arming);

	alarm = ALARM_OFF;
	led = LED_GREEN;
	return 0;
}

u8 arming() 
{
	TRANSITION( timer_c >= 250 , arming_lock);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 arming_lock() 
{
	TRANSITION( getRFIDState() == RFID_OFF , armed);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 armed() 
{
	TRANSITION( getRFIDState() == RFID_ON , disarming);
	TRANSITION( getDoorState() == DOOR_OPEN , trigger_stage_00);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 disarming() 
{
	TRANSITION( timer_c >= 250 , disarming_lock);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;

}

u8 disarming_lock() 
{
	TRANSITION( getRFIDState() == RFID_OFF , ready);

	alarm = ALARM_OFF;
	led = LED_GREEN;
	return 0;
}

u8 trigger_stage_00() 
{
	TRANSITION( getDoorState() == DOOR_CLOSED , armed);
	TRANSITION( timer_c >= 500 , trigger_stage_01);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_01() 
{
	TRANSITION( getDoorState() == DOOR_CLOSED , armed);
	TRANSITION( timer_c >= 250 , trigger_stage_02);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_02() 
{
	TRANSITION( getDoorState() == DOOR_CLOSED , armed);
	TRANSITION( timer_c >= 750 , trigger_stage_03);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_03() 
{
	TRANSITION( getDoorState() == DOOR_CLOSED , armed);
	TRANSITION( timer_c >= 250 , trigger_stage_04);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_04()
{
	TRANSITION( timer_c >= 250 , trigger_stage_05);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_05()
{
	TRANSITION( timer_c >= 250 , trigger_stage_06);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_06()
{
	TRANSITION( timer_c >= 750 , trigger_stage_07);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_07()
{
	TRANSITION( timer_c >= 250 , trigger_stage_08);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_08()
{
	TRANSITION( timer_c >= 250 , trigger_stage_09);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_09()
{
	TRANSITION( timer_c >= 250 , trigger_stage_10);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_10()
{
	TRANSITION( timer_c >= 250 , trigger_stage_11);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_11()
{
	TRANSITION( timer_c >= 250 , trigger_stage_12);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_12()
{
	TRANSITION( timer_c >= 750 , trigger_stage_13);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_13()
{
	TRANSITION( timer_c >= 30000 , trigger_stage_14);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_14()
{
	TRANSITION( timer_c >= 3000 , trigger_stage_15);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_15()
{
	TRANSITION( timer_c >= 45000 , trigger_stage_16);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_16()
{
	TRANSITION( timer_c >= 3000 , trigger_stage_17);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_17()
{
	TRANSITION( timer_c >= 60000 , trigger_stage_18);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_18()
{
	TRANSITION( timer_c >= 3000 , trigger_stage_19);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_19()
{
	TRANSITION( timer_c >= 60000 , trigger_stage_20);
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_ON;
	led = LED_RED;
	return 0;
}

u8 trigger_stage_20()
{
	TRANSITION( getRFIDState() == RFID_ON , disarming);

	alarm = ALARM_OFF;
	led = LED_RED;
	return 0;
}

/*** timer handler**********************************/

unsigned int blink_c = 0;
ISR (TIMER0_COMPA_vect)
{

	timer_c++;

	while (state() != 0) /* transition to next state */
	{
		timer_c = 0;
	}

	/* control horn */
	if ( (alarm == ALARM_ON) && 
	     (getAlarmOverrideState() == ALARM_ENABLED) 
	   ) {
		PORTB |= (1<<PB1);
	} else {
		PORTB &= ~(1<<PB1);
	}

	/* status led */
	if (led == LED_RED) {
		/* blink red/green */
		if (500 < (blink_c = (blink_c+1) % 1000)) 
		{
			PORTB |= (1<<PB2);
		} else {
			PORTB &= ~(1<<PB2);
		}
	} else {
		/* green on */
		PORTB &= ~(1<<PB2);
	}
}

main()
{
	int d;

	/* setup timer interrupt */
	TCCR0A |= (1<<WGM01);
	TCCR0B |= (1<<CS00) | (1<<CS01); /* 8000000/64 -> T=1/125000 */
	OCR0A = 124;	/* 125/125000 -> T=0.001ms */
	TIMSK |= (1<<OCIE0A);

	/* pull up hall door & disarm */
	PORTB = (1<<PB3) | (1<<PB4);
	/* out: LED & HORN */
	DDRB = (1<<PB2) | (1<<PB1); 

	/* give the system time to settle down */
	for(d = 2000; d>0; d--) 
	{
		_delay_ms(1);
	}

	sei();

        while(1)
        {
		/* nop */
        }
}

